﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
Life#Breathing v11 by MacGruber.
Audio-Synchronized Breathing for VaM.

Version 11 2020-11-19
	BreathingCore: Making "OnChooseBreath" available as VaM Trigger so you can connect to a LogicBrick/StateMachine.
	DriverThrustTrigger: New sub-plugin replacing DriverThrustSFX. Way more option for you by exposing to VaM Triggers. E.g. connect to LogicBrick/RandomSound or my Connect plugin.
	Improved UI: Getting rid of annoying scrollboxes around info panels.

Version 10 2020-08-23
	DriverBreathing: Reset duplicate morphs used by old versions to avoid causing references to multiple versions of Life.
	AudioAttenuation: Tweaked desktop default volume from 0.22 to 0.70.

Version 9 2020-08-19
	Fixed silent exception / log spam issue with DriverThrustSFX when not all SFX audio source are connected.

Version 8 2020-08-15
	Fixing "OnChooseBreath" broadcast throwing silent error when no other plugin is listening to the broadcast.
	Making sure morphs are reset to defaults when disabling the plugin. This is to avoid unnecessary dependencies to Life.

Version 7 2020-06-05
	Adding debug features for adding custom audio.
	Fixing rhythm damping setting in the Life01 scene.

Version 6 2020-03-27
	Added 36 new mouth breathing sounds!
	Implemented Orgasm support! Added 6 different orgasms to the AudioBundle, these are 3-8 chained breaths each.
	Implemented JSON scaling options to scale nose, mouth and lung animations independently for each breath.
	Implemented Multi-Breath support, these are multiple breaths in a single audio file.
	Implemented 'OnChooseBreath' broadcast message. This is essential for those who implement their own controller plugin to direct my Life plugin. It notifies you when it is about to choose the next breath and you should update settings (Intensity, ThrustMultiplier, etc.)
	Potential silent crash fixes for ThrustSFX.
	Updated Scene_Life03 controller plugin to support orgasms.

Version 5 2020-03-21
	Support for male characters, including Futa.
	Support for pitch, which controls audio playback speed. You can use this to cause frequency shifts, e.g. making audio sound slightly more male or older, etc.
	Support for custom Breathing-AudioBundles. Provide your own sounds!
	Shipping as VaM 1.19 package.

Version 0.4 2020-03-01
	Improved breathing-sync of Thrust plugin.
	Implemented SoundFX for Thrust plugin with Lube, FleshySlap and BedSqueak sounds.
	Implemented UpDown-Balance option for Thrust plugin in preparation for VaMConnect interface.
	Fixed audio compression for breathing audio.
	Forcing Pitch to 1.0 for breathing audio to prevent potential desyncs.
	Removed Volume seperation for Desktop/VR, use AudioDistanceAttenuation plugin instead.
	Added Life04 demo scene to show SFX.

Version 0.3.1 2020-02-16
	Fixed nose breathing-in offset.
	Fixed lips morph animating parts of nose.

Version 0.3 2020-02-15
	Implemented custom morphs for chest, stomach, nose (by DDQuidam)
	Implemented custom morph for lips (by TimelordToby)
	Fixed MacGruber_Life.cslist

Version 0.2 2020-01-20
	Implemented RhythmDamping to allow you control how regular you want your breathing.
	Implemented "DriverThrust", drives AnimationPatterns and/or VariableTriggers so you can have synchronized 'action'.
	Implemented separate volume for VR and Desktop.	Also rebalanced audio volume between recordings.
	Implemented audio-interrupt feature which allows faster transition between IO/OI modes.
	Implemented PostMagic-style info text to explain all the sliders.
	Fixed potential silent crash during scene load.
	Fixed issue with Breathing settings not surviving changing Person look.
	Tweaked some default values.

Version 0.1 2019-12-18
	Initial release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using UnityEngine.XR;
using MeshVR;
using System;
using System.Collections.Generic;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using SimpleJSON;

namespace MacGruber
{
	public partial class Breathing : MVRScript
	{
		////////////////////////////////////////////////////////////////////////////////////
		// Constants

		private static readonly string BASE_PATH = "Assets/Breathing/";

		public const int TAG_MOAN    = 1 << 0;
		public const int TAG_NOSEIN  = 1 << 1;
		public const int TAG_NOSEOUT = 1 << 2;
		public const sbyte PREV = 0;
		public const sbyte CURR = 1;
		public const sbyte NEXT = 2;
		public const int FORMAT_INOUT = 0;
		public const int FORMAT_OUTIN = 1;
		public const int TYPE_BREATH = 0;
		public const int TYPE_ORGASM = 1;
		public const int TYPE_DEBUG = 2;

		////////////////////////////////////////////////////////////////////////////////////
		// Members

		private string myBundleURL = null;
		private Request myBundleRequest = null;
		private JSONNode myDatabase;
		private Dictionary<string, Dataset> myDatasets = new Dictionary<string, Dataset>();
		private List<BreathEntry> mySelection = new List<BreathEntry>();
		private MiniQueue<Breath> myQueue = new MiniQueue<Breath>();
		private List<Event> myEvents = new List<Event>();
		private float myDurationDamped = -1.0f;
		private int myMultiBreathIndex = 0;
		private bool myInitialized = false;
		private bool myQueueInitialized = false;
		private int myQueueType = TYPE_BREATH;
		private string myQueueDebugDataset = "";
		private int myQueueDebugEntry = 0;

		private AudioSourceControl myHeadAudioSource;

		private JSONStorableUrl myAudioBundle;
		private JSONStorableStringChooser myBreathDataset;
		private JSONStorableStringChooser myOrgasmDataset;
		private JSONStorableFloat myPitch;
		private JSONStorableFloat myIntensity;
		private JSONStorableFloat myVariance;
		internal JSONStorableFloat myRhythmRandomness;
		internal JSONStorableFloat myRhythmDamping;
		private JSONStorableFloat myVolume;

		private EventTrigger myEventOnChooseBreath;

		public float CurrentTime { get; private set; }
		public float VelocityClampScale { get; private set; }

		////////////////////////////////////////////////////////////////////////////////////
		// Core

		public override void Init()
        {
			Cleanup();

			if (containingAtom.type != "Person")
			{
				SuperController.LogError("MacGruber Breathing script needs to be on a Person atom.");
				return;
			}

			InitCore();
			InitUI();
        }

		private void InitCore()
		{
			myHeadAudioSource = containingAtom.GetStorableByID("HeadAudioSource") as AudioSourceControl;

			RegisterCallbackEvent(SM(MRK_Interrupt), Interrupt);
			RegisterCallbackEvent(SM(MRK_BreathEnd), BreathEnd);
		}

		private void InitUI()
		{
			Utils.OnInitUI(CreateUIElement);
			Utils.SetupInfoTextNoScroll(this,
				"<color=#606060><size=40><b>Breathing</b></size>\nThis selects and plays breathing audio as well as providing the event system as a basis for animation. Animation driver plugins can connect to this to control their body animation.</color>\n\n" +
				"<b>AudioBundle:</b> Select an <i>AudioBundle</i> with breathing audio and metadata to load.\n\n" +
				"<b>Breath Dataset:</b> Choose between available audio recording sets within the selected <i>AudioBundle</i>.\n\n" +
				"<b>Orgasm Dataset:</b> Choose between available audio recording sets within the selected <i>AudioBundle</i>.\n\n" +
				"<b>Queue Orgasm:</b> Queue one orgasm from the selected Orgasm dataset. This button is mostly for testing, normally you would trigger this from some other plugin or via trigger action.\n\n" +
				"<b>Pitch:</b> Playback speed of audio files. Use this to cause frequency shifts, e.g. making audio sound more male.\n\n" +
				"<b>Intensity:</b> Intensity of breathing. Higher values generally mean faster breathing.\n\n" +
				"<b>Audio Variance:</b> Variance when selecting audio files. Higher variance means less repetition of the same audio, but higher variance in intensity.\n\n" +
				"<b>Rhythm Randomness:</b> Artificial randomness on wait time between breaths. Use as alternative to 'Audio Variance'.\n\n" +
				"<b>Rhythm Damping:</b> Damping on duration of each breath. This is achieved by adjusting wait time between breaths. Higher values mean a more regular breathing, but slower adjustment to different intensities.\n\n" +
				"<b>Volume:</b> Volume of the <i>AudioSource</i>. Note that this plugin continuously overrides the <i>HeadAudio</i> volume of your <i>Person</i> atom.\n\n" +
				"<b>OnChooseBreath:</b> Setup actions right before a new breath is choosen. This is the optimal time for tweaking Life settings. Common example would be to call <i>TriggerSync</i> on a <i>StateMachine</i> out of the <i>LogicBricks</i> plugin collection.\n\n",
				1950.0f, true
			);

			string defaultBundle = Utils.GetPluginPath(this)+"/MacGruber_Breathing.audiobundle";
			myAudioBundle = Utils.SetupAssetBundleChooser(this, "AudioBundle", defaultBundle, false, "audiobundle");
			myAudioBundle.setCallbackFunction += (string url) => {
				if (myBundleURL != null)
					AssetLoader.DoneWithAssetBundleFromFile(myBundleURL);
				myBundleURL = url;
				//SuperController.LogMessage(myBundleURL);
				Request request = new AssetLoader.AssetBundleFromFileRequest {path = myBundleURL, callback = OnAssetBundleLoaded};
				AssetLoader.QueueLoadAssetBundleFromFile(request);
			};

			myBreathDataset = new JSONStorableStringChooser("Breath Dataset", new List<string>(), string.Empty, "Breath Dataset");
			myBreathDataset.storeType = JSONStorableParam.StoreType.Full;
			CreateScrollablePopup(myBreathDataset, false);
			RegisterStringChooser(myBreathDataset);

			myOrgasmDataset = new JSONStorableStringChooser("Orgasm Dataset", new List<string>(), string.Empty, "Orgasm Dataset");
			myOrgasmDataset.storeType = JSONStorableParam.StoreType.Full;
			CreateScrollablePopup(myOrgasmDataset, false);
			RegisterStringChooser(myOrgasmDataset);

			Utils.SetupButton(this, "Queue Orgasm", QueueOrgasm, false);
			Utils.SetupAction(this, "QueueOrgasm", QueueOrgasm);

			myPitch = Utils.SetupSliderFloat(this, "Pitch", 1.0f, 0.5f, 2.0f, false);
			myIntensity = Utils.SetupSliderFloat(this, "Intensity", 0.0f, 0.0f, 1.0f, false);
			myVariance = Utils.SetupSliderFloat(this, "Audio Variance", 8.0f, 5.0f, 12.0f, false);
			myVariance.setCallbackFunction += (float v) => {
				float r = Mathf.Round(v);
				if (myVariance.val != r)
					myVariance.val = r;
			};
			myRhythmRandomness = Utils.SetupSliderFloat(this, "Rhythm Randomness", 0.07f, 0.0f, 0.2f, false);
			myRhythmDamping = Utils.SetupSliderFloat(this, "Rhythm Damping", 0.3f, 0.0f, 0.9f, false);

			myVolume = Utils.SetupSliderFloat(this, "Volume", 0.5f, 0.0f, 1.0f, false);

			myEventOnChooseBreath = new EventTrigger(this, "OnChooseBreath");
			Utils.SetupButton(this, "Actions OnChooseBreath", myEventOnChooseBreath.OpenPanel, false);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void OnAtomRename(string oldid, string newid)
		{
			myEventOnChooseBreath.SyncAtomNames();
		}

		private void OnAssetBundleLoaded(Request aRequest)
		{
			Cleanup();
			myBundleRequest = aRequest;

			TextAsset ta = myBundleRequest.assetBundle?.LoadAsset<TextAsset>(BASE_PATH + "Database.json");
			if (ta == null)
			{
				SuperController.LogError("Failed loading JSON database from AssetBundle.");
				return;
			}

			try
			{
				JSONNode database = JSON.Parse(ta.text);
				LoadDatabase(database);
			}
			catch (Exception e)
			{
				SuperController.LogError("MacGruber Breathing: Exception while parsing JSON: " + e);
			}
		}

		private void OnDestroy()
        {
			Cleanup();
            if (myBundleURL != null)
                AssetLoader.DoneWithAssetBundleFromFile(myBundleURL);
			myBundleURL = null;
			myEvents.Clear();

			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			myEventOnChooseBreath.Remove();

			Utils.OnDestroyUI();
        }

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[myEventOnChooseBreath.Name] = myEventOnChooseBreath.GetJSON(base.subScenePrefix);
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				myEventOnChooseBreath.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}

		private void Cleanup()
		{
			myInitialized = false;
			myQueueInitialized = false;
			myQueue.Clear();
			myDatasets.Clear();
			mySelection.Clear();
			myDurationDamped = -1;
			myMultiBreathIndex = 0;
		}

		private void LoadDatabase(JSONNode aDatabase)
		{
			myDatabase = aDatabase;
			if (myDatabase == null || myDatabase["datasets"] == null)
			{
				SuperController.LogError("Failed parsing JSON database.");
				return;
			}

			List<string> breathChoices = new List<string>();
			List<string> orgasmChoices = new List<string>();
			JSONClass datasets = myDatabase["datasets"].AsObject;
			IEnumerator<string> datasetNames = datasets.Keys.GetEnumerator();
			while (datasetNames.MoveNext())
			{
				Dataset dataset = LoadDataset(datasetNames.Current);
				if (dataset == null)
					continue;
				if (dataset.myType == TYPE_BREATH)
					breathChoices.Add(dataset.myName);
				else if (dataset.myType == TYPE_ORGASM)
					orgasmChoices.Add(dataset.myName);
			}

			myBreathDataset.choices = breathChoices;
			if (breathChoices.Count > 0 && !breathChoices.Contains(myBreathDataset.val))
				myBreathDataset.val = breathChoices[0];

			myOrgasmDataset.choices = orgasmChoices;
			if (orgasmChoices.Count > 0 && !orgasmChoices.Contains(myOrgasmDataset.val))
				myOrgasmDataset.val = orgasmChoices[0];
		}

		private Dataset LoadDataset(string aDatasetName)
		{
			JSONNode datasetNode = myDatabase["datasets"]?[aDatasetName];
			if (datasetNode == null)
			{
				SuperController.LogError("Failed loading '"+aDatasetName+"' dataset.");
				return null;
			}

			string path = datasetNode["path"];
			string settypeStr = datasetNode["type"];
			int settype = TYPE_BREATH;
			if (settypeStr == "orgasm")
				settype = TYPE_ORGASM;

			string formatStr = datasetNode["format"];
			int format = FORMAT_INOUT;
			if (formatStr == "outin")
				format = FORMAT_OUTIN;
			JSONNode jsonVolume = datasetNode["volume"];
			float volume = jsonVolume != null ? datasetNode["volume"].AsFloat : 1.0f;

			JSONArray dataentries = datasetNode["entries"].AsArray;
			Dataset dataset = new Dataset(aDatasetName, settype);
			dataset.myEntries = new List<BreathEntry>(dataentries.Count);
			myDatasets[aDatasetName] = dataset;
			bool skippedEntries = false;
			for (int i=0; i<dataentries.Count; ++i)
			{
				BreathEntry entry = new BreathEntry(myBundleRequest, path, format, volume, dataentries[i]);
				if (entry.IsValid())
					dataset.myEntries.Add(entry);
				else
					skippedEntries = true;
			}
			if (skippedEntries)
				SuperController.LogError("Skipped invalid entries when loading dataset '"+aDatasetName+"'.");

			//SuperController.LogMessage("Loaded dataset '"+aDatasetName+"'.");
			myInitialized |= dataset.myEntries.Count > 0;

			return dataset;
		}

		private void Update()
		{
			myEventOnChooseBreath.Update();
			
			if (SuperController.singleton.isLoading)
				return;
			if (myBundleURL == null)
				myAudioBundle.setCallbackFunction(myAudioBundle.val);
			if (!myInitialized)
				return;

			float pitch = myPitch.val;
			if (pitch != myHeadAudioSource.pitch)
				myHeadAudioSource.pitch = pitch;

			if (myQueueInitialized)
			{
				CurrentTime += Time.deltaTime * pitch;
			}
			else
			{
				myQueue.Enqueue(new Breath());
				myQueue.Enqueue(new Breath() { Duration = UnityEngine.Random.Range(0.5f, 1.5f) });
				myQueue.Enqueue(new Breath());
				UpdateBreathTime();
				Choose();
				Breath currentBreath = Get(CURR);
				CurrentTime = currentBreath.Start;
				VelocityClampScale = currentBreath.VCS;
				myQueueInitialized = true;
				for (int i=0; i<myEvents.Count; ++i)
					myEvents[i].Reset();
			}

			for (int i=0; i<myEvents.Count; ++i)
				myEvents[i].Update(this);

			if (CurrentTime >= Get(NEXT).Start)
				Enqueue();
		}

		private void Interrupt()
		{
			Choose();
			Breath curr = Get(CURR);
			Breath next = Get(NEXT);

			bool forceInterrupt = curr.IsValid() && next.IsValid() && curr.Entry.Format != next.Entry.Format;
			if (forceInterrupt)
				curr.InvalidTime = CurrentTime - curr.Start;
			for (int i=0; i<myEvents.Count; ++i)
				myEvents[i].Invalidate(this);

			if (forceInterrupt)
			{
				CurrentTime = next.Start;
				Enqueue();
			}
		}

		private void BreathEnd()
		{
			if (CurrentTime < Get(NEXT).Start)
			{
				// We are NOT about to Enqueue the next breath. Therefore this is a multi-breath, we need to update data.
				++myMultiBreathIndex;
				Breath currentBreath = Get(CURR);
				currentBreath.UpdateScales(myMultiBreathIndex);
				VelocityClampScale = currentBreath.VCS;
			}
		}

		private void Enqueue()
		{
			CurrentTime -= Get(PREV).Duration;
			myQueue.Enqueue(new Breath());
			UpdateBreathTime();
			Breath currentBreath = Get(CURR);
			VelocityClampScale = currentBreath.VCS;
			if (myQueueType != TYPE_BREATH && currentBreath.Type != TYPE_BREATH)
				myQueueType = TYPE_BREATH;
			Choose();

			for (int i=0; i<myEvents.Count; ++i)
				myEvents[i].Enqueue(this);

			BreathEntry currentEntry = currentBreath.Entry;
			myHeadAudioSource.volume = Mathf.Clamp01(myVolume.val * currentEntry.Volume);
			myHeadAudioSource.PlayNow(currentEntry.Clip);
			myMultiBreathIndex = 0;

			//SuperController.LogMessage("Playing " + currentEntry.Name);
		}

		private void Choose()
		{
			myEventOnChooseBreath.Trigger();

			if (myQueueType == TYPE_BREATH)
				ChooseBreath();
			else if (myQueueType == TYPE_ORGASM)
				ChooseOrgasm();
			else if (myQueueType == TYPE_DEBUG)
				ChooseDebug();
			else
				ChooseBreath();
		}

		private void ChooseBreath()
		{
			Breath currentBreath = Get(CURR);
			Breath nextBreath = Get(NEXT);
			nextBreath.Type = TYPE_BREATH;
			nextBreath.Start = currentBreath.Start + currentBreath.Duration;

			Dataset dataset;
			if (myDatasets.TryGetValue(myBreathDataset.val, out dataset))
			{
				List<BreathEntry> entries = dataset.myEntries;

				float variance = myVariance.val * 0.5f;
				float intensity = Mathf.Lerp(variance-0.5f, entries.Count-variance-0.5f, myIntensity.val);
				float fmin = intensity - variance;
				float fmax = intensity + variance;
				int min = Mathf.RoundToInt(fmin);
				int max = Mathf.RoundToInt(fmax);
				if (UnityEngine.Random.value > min + 0.5f - fmin)
					min++;
				if (UnityEngine.Random.value > fmax - max + 0.5f)
					max--;
				min = Mathf.Clamp(min, 0, entries.Count-1);
				max = Mathf.Clamp(max, 0, entries.Count-1);

				mySelection.Clear();
				BreathEntry curr = Get(CURR).Entry;
				BreathEntry prev = Get(PREV).Entry;
				bool hasCurr = false;
				bool hasPrev = false;
				for (int i=min; i<=max; ++i)
				{
					BreathEntry entry = entries[i];
					if      (curr == entry) { hasCurr = true; continue; }
					else if (prev == entry) { hasPrev = true; continue; }
					mySelection.Add(entry);
				}

				// Fallbacks in case there are not enough nextBreaths to select from. Allow some more repetition
				if (mySelection.Count < 2 && hasPrev)
					mySelection.Add(prev);
				if (mySelection.Count < 1 && hasCurr)
					mySelection.Add(curr);

				if (mySelection.Count > 0)
				{
					int selectionIdx = UnityEngine.Random.Range(0, mySelection.Count);
					BreathEntry entry = mySelection[selectionIdx];
					float nextBreathStart = entry.GetFileLastBreathTime();
					float nextBreathEnd = entry.GetFileEndTime();
					float nextFile = entry.GetFileNextTime();
					if (nextBreathStart < 0)
						nextBreathStart = 0.0f;
					if (nextBreathEnd < 0)
						nextBreathEnd = 1.0f;
					if (nextBreathStart > nextBreathEnd-0.1f)
						nextBreathStart = Mathf.Max(0.0f, nextBreathEnd-0.1f);
					if (nextFile < 0)
						nextFile = nextBreathEnd + 0.1f;
					else if (nextFile < nextBreathEnd)
						nextFile = nextBreathEnd;

					float wait = nextFile - nextBreathEnd;
					float r = myRhythmRandomness.val;
					float duration = nextBreathEnd - nextBreathStart + Mathf.Max(wait * UnityEngine.Random.Range(1-r, 1+r), 0.05f);
					if (myDurationDamped > 0)
						myDurationDamped = Mathf.Lerp(duration, myDurationDamped, myRhythmDamping.val);
					else
						myDurationDamped = duration;

					nextBreath.Duration = Mathf.Max(myDurationDamped + nextBreathStart, nextBreathEnd + 0.05f);
					nextBreath.Entry = entry;
					nextBreath.Intensity = myIntensity.val;
					nextBreath.UpdateScales(0);
				}
			}
		}

		private void ChooseOrgasm()
		{
			Breath currentBreath = Get(CURR);
			Breath nextBreath = Get(NEXT);
			nextBreath.Type = TYPE_ORGASM;
			nextBreath.Start = currentBreath.Start + currentBreath.Duration;

			Dataset dataset;
			if (myDatasets.TryGetValue(myOrgasmDataset.val, out dataset))
			{
				List<BreathEntry> entries = dataset.myEntries;
				if (entries.Count > 0)
				{
					int selectionIdx = UnityEngine.Random.Range(0, entries.Count);
					BreathEntry entry = entries[selectionIdx];
					float nextBreathStart = entry.GetFileLastBreathTime();
					float nextBreathEnd = entry.GetFileEndTime();
					float nextFile = entry.GetFileNextTime();
					if (nextBreathStart < 0)
						nextBreathStart = 0.0f;
					if (nextBreathEnd < 0)
						nextBreathEnd = 1.0f;
					if (nextBreathStart > nextBreathEnd-0.1f)
						nextBreathStart = Mathf.Max(0.0f, nextBreathEnd-0.1f);
					if (nextFile < 0)
						nextFile = nextBreathEnd + 0.1f;
					else if (nextFile < nextBreathEnd)
						nextFile = nextBreathEnd;

					float wait = nextFile - nextBreathEnd;
					float r = myRhythmRandomness.val;
					float duration = nextBreathEnd + Mathf.Max(wait * UnityEngine.Random.Range(1-r, 1+r), 0.05f);
					myDurationDamped = duration - nextBreathStart;
					nextBreath.Duration = duration;
					nextBreath.Entry = entry;
					nextBreath.Intensity = myIntensity.val;
					nextBreath.UpdateScales(0);
				}
			}
		}

		private void ChooseDebug()
		{
			Breath currentBreath = Get(CURR);
			Breath nextBreath = Get(NEXT);
			nextBreath.Type = TYPE_DEBUG;
			nextBreath.Start = currentBreath.Start + currentBreath.Duration;

			Dataset dataset;
			if (myDatasets.TryGetValue(myQueueDebugDataset, out dataset))
			{
				List<BreathEntry> entries = dataset.myEntries;
				if (myQueueDebugEntry >= 0 && myQueueDebugEntry < entries.Count)
				{
					BreathEntry entry = entries[myQueueDebugEntry];
					float nextBreathStart = entry.GetFileLastBreathTime();
					float nextBreathEnd = entry.GetFileEndTime();
					float nextFile = entry.GetFileNextTime();
					if (nextBreathStart < 0)
						nextBreathStart = 0.0f;
					if (nextBreathEnd < 0)
						nextBreathEnd = 1.0f;
					if (nextBreathStart > nextBreathEnd-0.1f)
						nextBreathStart = Mathf.Max(0.0f, nextBreathEnd-0.1f);
					if (nextFile < 0)
						nextFile = nextBreathEnd + 0.1f;
					else if (nextFile < nextBreathEnd)
						nextFile = nextBreathEnd;

					float wait = nextFile - nextBreathEnd;
					float r = myRhythmRandomness.val;
					float duration = nextBreathEnd + Mathf.Max(wait * UnityEngine.Random.Range(1-r, 1+r), 0.05f);
					myDurationDamped = duration - nextBreathStart;
					nextBreath.Duration = duration;
					nextBreath.Entry = entry;
					nextBreath.Intensity = myIntensity.val;
					nextBreath.UpdateScales(0);
				}
			}
		}

		private void UpdateBreathTime()
		{
			float time = 0.0f;
			for (int i=0; i<GetCapacity(); ++i)
			{
				Breath b = myQueue.Get(i);
				b.Start = time;
				time += b.Duration;
			}
		}

		public void QueueOrgasm()
		{
			myQueueType = TYPE_ORGASM;
		}


		////////////////////////////////////////////////////////////////////////////////////
		// Helpers

		public Breath Get(int index)
		{
			return myQueue.Get(index);
		}

		public int GetCapacity() { return MiniQueue<Breath>.Capacity; }

		/*private void FullLog()
		{
			string msg = "FullLog: " + CurrentTime + "\n";
			for (int i=0; i<GetCapacity(); ++i)
			{
				Breath b = myQueue.Get(i);
				msg += "[" + i + "] ## Start=" + b.Start + ", Duration=" + b.Duration + ", End=" + (b.Start+b.Duration) + ", Invalid=" + b.InvalidTime + "\n";
				if (b.IsValid())
				{
					msg += "    ## ";
					int len = msg.Length;
					for (int j=0; j<b.Entry.Markers.Count; ++j)
					{
						BreathMarker m = b.Entry.Markers[j];
						if (m.NameHash == MRK_BreathIn)
							msg += "MRK_BreathIn=" + (b.Start+m.Time);
						else if (m.NameHash == MRK_HoldIn)
							msg += "MRK_HoldIn=" + (b.Start+m.Time);
						else if (m.NameHash == MRK_BreathOut)
							msg += "MRK_BreathOut=" + (b.Start+m.Time);
						else if (m.NameHash == MRK_HoldOut)
							msg += "MRK_HoldOut=" + (b.Start+m.Time);
						else if (m.NameHash == MRK_Interrupt)
							msg += "MRK_Interrupt=" + (b.Start+m.Time);
						else if (m.NameHash == MRK_BreathEnd)
							msg += "MRK_BreathEnd=" + (b.Start+m.Time);
						else
							continue;

						msg += ", ";
						if (m.NameHash == MRK_BreathEnd)
							msg += "\n    ## ";
					}
					msg += "\n";
				}
			}
			SuperController.LogMessage(msg);
		}*/

		////////////////////////////////////////////////////////////////////////////////////
		// Datastructures

		private class MiniQueue<T> where T : new()
		{
			public const int Capacity = 3;

			private int position = 0;
			private T[] data = new T[Capacity];

			public T Get(int index) {
				return data[(position + index) % Capacity];
			}

			public void Enqueue(T t)
			{
				data[position] = t;
				position = (position+1) % Capacity;
			}

			public void Clear()
			{
				for (int i=0; i<Capacity; ++i)
					data[i] = new T();
				position = 0;
			}
		}

		public class Breath
		{
			public BreathEntry Entry;
			public int Type = TYPE_BREATH;
			public float Start;
			public float Duration = 1.0f;
			public float Intensity = 0.0f;
			public float Power = 1.0f;
			public float Depth = 1.0f;
			public float Lung = 1.0f;
			public float Mouth = 1.0f;
			public float Nose = 1.0f;
			public float VCS = 1.0f;
			public float InvalidTime = float.MaxValue;

			public bool IsValid()
			{
				return Entry != null;
			}

			public void UpdateScales(int aMultiBreathIndex)
			{
				Depth = Entry.Depth(aMultiBreathIndex);
				if (Depth < 0)
					Depth = Intensity*0.5f+0.5f;
				Lung = Entry.Lung(aMultiBreathIndex);
				Mouth = Entry.Mouth(aMultiBreathIndex);
				Nose = Entry.Nose(aMultiBreathIndex);
				VCS = Entry.VCS(aMultiBreathIndex);
				Power = UnityEngine.Random.Range(0.8f, 1.2f);
			}
		}

		public class BreathEntry
		{
			public AssetBundleAudioClip Clip { get; private set; }
			public string Name { get; private set; }
			public int Tags { get; private set; }
			public int Format { get; private set; }
			public float Volume { get; private set; }
			public List<BreathMarker> Markers { get; private set; }

			public float Depth(int i) {	return myDepth[Mathf.Clamp(i, 0, myDepth.Length-1)]; }
			public float Lung(int i)  { return myLung [Mathf.Clamp(i, 0, myLung.Length-1) ]; }
			public float Mouth(int i) {	return myMouth[Mathf.Clamp(i, 0, myMouth.Length-1)]; }
			public float Nose(int i)  { return myNose [Mathf.Clamp(i, 0, myNose.Length-1) ]; }
			public float VCS(int i)   { return myVCS  [Mathf.Clamp(i, 0, myVCS.Length-1)  ]; }

			private float[] myDepth;
			private float[] myLung;
			private float[] myMouth;
			private float[] myNose;
			private float[] myVCS;

			public BreathEntry(Request aRequest, string aPath, int aFormat, float aVolume, JSONNode aEntryData)
			{
				Name = aPath + aEntryData["file"].Value;
				Format = aFormat;
				Volume = aVolume;

				Tags = 0;
				string tags = aEntryData["tags"].Value;
				if (tags != null)
				{
					string[] tagsList = tags.Split(',');
					if (Array.Exists(tagsList, t => t == "moan"))
						Tags |= TAG_MOAN;
					if (Array.Exists(tagsList, t => t == "noseIn"))
						Tags |= TAG_NOSEIN;
					if (Array.Exists(tagsList, t => t == "noseOut"))
						Tags |= TAG_NOSEOUT;
				}

				JSONArray markers = aEntryData["markers"].AsArray;
				Markers = new List<BreathMarker>(markers.Count+2);
				BreathMarker m;
				for (int i=0; i<markers.Count; ++i)
				{
					JSONArray a = markers[i].AsArray;
					m.NameHash = a[0].Value.GetHashCode();
					m.Time = a[1].AsFloat;
					Markers.Add(m);
				}

				m.NameHash = MRK_Start;
				m.Time = 0.0f;
				Markers.Add(m);
				m.NameHash = MRK_End;
				m.Time = float.MaxValue; // will be overridden with the duration of the entry later (in MarkerImpl.Update)
				Markers.Add(m);

				Markers.Sort();

				{ // Add BreathEnd markers between breaths of multi-breath entries
					int markerA = (aFormat == FORMAT_OUTIN) ? MRK_HoldIn    : MRK_HoldOut;
					int markerB = (aFormat == FORMAT_OUTIN) ? MRK_BreathOut : MRK_BreathIn;
					for (int i=0; i<Markers.Count-1; ++i)
					{
						if (Markers[i].NameHash != markerA)
							continue;
						int j = Markers.FindIndex(i+1, bm => bm.NameHash == markerB);
						if (j < 0)
							break;
						m.NameHash = MRK_BreathEnd;
						m.Time = (Markers[i].Time + Markers[j].Time) * 0.5f;
						Markers.Add(m);
					}
				}

				{ // Add interrupt marker in the middle of the last breath in the entry
					int markerA = (aFormat == FORMAT_OUTIN) ? MRK_HoldOut : MRK_HoldIn;
					int markerB = (aFormat == FORMAT_OUTIN) ? MRK_BreathIn : MRK_BreathOut;
					int i = Markers.FindLastIndex(bm => bm.NameHash == markerA);
					int j = Markers.FindLastIndex(bm => bm.NameHash == markerB);
					if (i >= 0 && j >= 0)
					{
						m.NameHash = MRK_Interrupt;
						m.Time = (Markers[i].Time + Markers[j].Time) * 0.5f;
						Markers.Add(m);
					}
				}

				{ // Sanity check
					ourSanityCounts.Clear();
					for (int i=0; i<Markers.Count; ++i)
					{
						int count = 1;
						int hash = Markers[i].NameHash;
						if (ourSanityCounts.TryGetValue(hash, out count))
							++count;
						else
							count = 1;
						ourSanityCounts[hash] = count;
					}

					int breathEndCount, breathOutCount, holdOutCount, breathInCount, holdInCount;
					ourSanityCounts.TryGetValue(MRK_BreathEnd, out breathEndCount);
					ourSanityCounts.TryGetValue(MRK_BreathOut, out breathOutCount);
					ourSanityCounts.TryGetValue(MRK_HoldOut, out holdOutCount);
					ourSanityCounts.TryGetValue(MRK_BreathIn, out breathInCount);
					ourSanityCounts.TryGetValue(MRK_HoldIn, out holdInCount);

					if (breathEndCount <= 0 || breathOutCount != breathEndCount || holdOutCount != breathEndCount ||
						breathInCount  != breathEndCount ||	holdInCount    != breathEndCount)
					{
						SuperController.LogError("MacGruber Breathing: Number of markers does not match detected number of breaths! Check your markers for file '"+BASE_PATH+Name+"'. " +
							"Order of markers needs to match time index and format. Note that accidental use of a comma symbol instead of a point symbol for float numbers can cause unintended reordering. "+
							"Breaths detected: "+breathEndCount+", Markers detected: BreathOut="+breathOutCount+
						    ", HoldOut="+holdOutCount+", BreathIn="+breathInCount+", HoldIn="+holdInCount);
					}
				}

				Markers.Sort();

				myDepth = GetFloats(aEntryData["depth"], -1.0f);
				myLung  = GetFloats(aEntryData["lung"],   1.0f);
				myMouth = GetFloats(aEntryData["mouth"],  1.0f);
				myNose  = GetFloats(aEntryData["nose"],   1.0f);
				myVCS   = GetFloats(aEntryData["vcs"],    1.0f);

				Clip = new AssetBundleAudioClip(aRequest, BASE_PATH, Name);
			}

			private static Dictionary<int, int> ourSanityCounts = new Dictionary<int,int>();

			private float[] GetFloats(JSONNode aNode, float aDefault)
			{
				if (aNode == null)
					return new float[] { aDefault };

				JSONArray array = aNode.AsArray;
				if (array != null)
				{
					if (array.Count == 0)
						return new float[] { aDefault };

					float[] numbers = new float[array.Count];
					for (int i=0; i<array.Count; ++i)
						numbers[i] = array[i].AsFloat;
					return numbers;
				}

				return new float[] { aNode.AsFloat };
			}

			public bool IsValid()
			{
				return (Clip?.sourceClip) != null;
			}

			public bool HasTag(int aTag)
			{
				return (Tags & aTag) > 0;
			}

			public float GetFileEndTime()
			{
				int idx = Markers.FindLastIndex(m => m.NameHash == MRK_BreathEnd);
				if (idx >= 0)
					return Markers[idx].Time;
				else
					return -1.0f;
			}

			public float GetFileLastBreathTime()
			{
				int marker = (Format == FORMAT_OUTIN) ? MRK_BreathOut : MRK_BreathIn;
				int idx = Markers.FindLastIndex(m => m.NameHash == marker);
				if (idx >= 0)
					return Markers[idx].Time;
				else
					return -1.0f;
			}

			public float GetFileNextTime()
			{
				int idx = Markers.FindLastIndex(m => m.NameHash == MRK_FileNext);
				if (idx >= 0)
					return Markers[idx].Time;
				else
					return -1.0f;
			}
		}

		public struct BreathMarker : IComparable<BreathMarker>
		{
			public int NameHash;
			public float Time;

			public int CompareTo(BreathMarker aOther)
			{
				return Time.CompareTo(aOther.Time);
			}
		}

		public class Dataset
		{
			public string myName;
			public int myType;
			public List<BreathEntry> myEntries;

			public Dataset(string aName, int aType)
			{
				myName = aName;
				myType = aType;
			}
		}
	}
}