﻿using UnityEngine;
using System;
using System.Collections.Generic;
using SimpleJSON;
using static MacGruber.Breathing;

namespace MacGruber
{
	public class DriverThrust : MVRScript
	{
		private Breathing myBreathing;
		private JSONStorableBool mySmooth;
		private JSONStorableBool myAPCompensation;
		private JSONStorableFloat myPower;
		private JSONStorableFloat myMultiplier;
		private JSONStorableFloat mySpeedDamping;
		private JSONStorableFloat myOffset;
		private JSONStorableFloat myUpDownBalance;
        private FloatTrigger myFloatTrigger;
		private ThrustEvent myThrustEvent;
		private bool myEventInit = false;
		private bool myWasLoading = true;

		private float myIncrement = 0.5f;
		private float myDirection = 1.0f;
		private float myPowerScale = 1.0f;
		private float myValue = -1;
		private float myEventUp = 0;
		private float myEventDown = 0;
		private int myMultiplierValue;

		public float Value { get { return myValue; } }
		public float DurationUp { get; private set; }
		public float DurationDown { get; private set; }
		public float Power { get { return myPower.val; } }

		public override void Init()
		{
			myBreathing = Utils.FindWithinSamePlugin<Breathing>(this);
			InitUI();
			InitAnimation();
			InitEvents();

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void InitUI()
		{
			Utils.OnInitUI(CreateUIElement);
			Utils.SetupInfoTextNoScroll(this,
				"<color=#606060><size=40><b>DriverThrust</b></size>\nThis animation driver controls <i>AnimationPatterns</i> or <i>VariableTriggers</i> from breathing. " +
				"The plugin is searching for atoms whose name starts with 'Person#Thrust', where 'Person' is the name of your <i>Person</i> atom. You can control as many <i>AnimationPatterns</i> or <i>VariableTriggers</i> as you want.\n" +
				"The animation should range from 0 to 1, where 0 is the 'inside' position. The plugin tries continuously adjust speed to " +
				"keep 'going in when breathing out'.</color>\n\n" +
				"<b>Power:</b> Range of animation being used. E.g. a value of 0.8 means only the range from 0.0 to 0.8 is used. Note that even with a power value of the 1.0 a part of the available range is used to smooth out animation velocity when breaths are not precisely regular.\n\n" +
				"<b>Smooth:</b> Apply classic <i>SmoothStep</i> to output value instead of using linear motion.\n\n" +
				"<b>AP Compensation:</b> Compensate for VaM <i>AnimationPatterns</i> using cubic bezier curves for interpolation, making the motion more linear. Recommended when driving <i>AnimationPatterns</i>.\n\n" +
				"<b>Multiplier:</b> Number of thrusts per breath. Higher values are recommended to be used with high 'Rhythm Damping' in the main plugin.\n\n" +
				"<b>Offset:</b> Sync offset in seconds between animation and audio. Generally you want to start the animation slightly earlier. Less so when using Multiplier.\n\n" +
				"<b>UpDownBalance:</b> When using suction devices it can be useful that it should suck slightly longer than blow, to prevent it falling off. Value in seconds, positive values mean going down takes longer.\n\n" +
				"<b>Rhythm Randomness:</b> Artificial randomness on wait time between breaths. <color=#606060>Slider linked from Breathing plugin for convenience.</color>\n\n" +
				"<b>Rhythm Damping:</b> Damping on duration of each breath. Increase if you have trouble with syncing thrust animation to audio. <color=#606060>Slider linked from Breathing plugin for convenience.</color>\n\n",
				1900.0f, true
			);

			myFloatTrigger = new FloatTrigger(this, "OnValueSet");
			Utils.SetupButton(this, "Actions OnValueSet", myFloatTrigger.OpenPanel, false);

			myPower = Utils.SetupSliderFloat(this, "Power", 1.0f, 0.0f, 1.0f, false);
			mySmooth = Utils.SetupToggle(this, "Smooth", false, false);
			myAPCompensation = Utils.SetupToggle(this, "AP Compensation", true, false);
			myMultiplier = Utils.SetupSliderFloat(this, "Multiplier", 1.0f, 1.0f, 5.0f, false);
			myMultiplier.setCallbackFunction  += (float v) => {
				myMultiplier.valNoCallback = Mathf.Round(v);
			};

			myOffset = Utils.SetupSliderFloat(this, "Offset", -0.1f, -0.3f, 0.1f, false);
			myUpDownBalance = Utils.SetupSliderFloat(this, "UpDownBalance", 0.0f, -0.5f, 0.5f, false);


			JSONStorableFloat rrm = myBreathing.myRhythmRandomness;
			JSONStorableFloat rrl = Utils.SetupSliderFloat(this, rrm.name, rrm.defaultVal, rrm.min, rrm.max, false);
			rrm.setCallbackFunction += (float v) => { rrl.valNoCallback = v; };
			rrl.setCallbackFunction += (float v) => { rrm.val = v; };
			rrl.valNoCallback = rrm.val;

			JSONStorableFloat rdm = myBreathing.myRhythmDamping;
			JSONStorableFloat rdl = Utils.SetupSliderFloat(this, rdm.name, rdm.defaultVal, rdm.min, rdm.max, false);
			rdm.setCallbackFunction += (float v) => { rdl.valNoCallback = v; };
			rdl.setCallbackFunction += (float v) => { rdm.val = v; };
			rdl.valNoCallback = rdm.val;
        }

		private void InitAnimation()
		{
			myValue = 0.0f;
			myMultiplierValue = Mathf.RoundToInt(myMultiplier.val);
			myIncrement = 2.0f * myMultiplier.val / 3.0f;
			myDirection = 1.0f;
			myPowerScale = 1.0f;
		}

		private void InitEvents()
		{
			myThrustEvent = new ThrustEvent(this);
			myBreathing.RegisterEvent(myThrustEvent);

			myOffset.setCallbackFunction  += (float v) => {
				myThrustEvent.Offset = v;
			};
			myOffset.setCallbackFunction(myOffset.val);
			myEventInit = true;
		}

		private void OnEnable()
		{
			if (myEventInit)
			{
				myBreathing.UnregisterEvent(myThrustEvent);
				myBreathing.RegisterEvent(myThrustEvent);
			}
		}

		private void OnDisable()
		{
			if (myEventInit)
				myBreathing.UnregisterEvent(myThrustEvent);
			myWasLoading = true;
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

			if (myFloatTrigger != null)
				myFloatTrigger.Remove();

			Utils.OnDestroyUI();
		}

		private void OnAtomRename(string oldid, string newid)
		{
			myFloatTrigger.SyncAtomNames();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[myFloatTrigger.Name] = myFloatTrigger.GetJSON(base.subScenePrefix);
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				myFloatTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}

		private void ThrustUpdate()
		{
			float duration = 1.0f / myIncrement;
			float balanceUp = Mathf.Min(myUpDownBalance.val * 0.5f, Mathf.Max(duration-0.05f, 0.0f));
			float balanceDown = myUpDownBalance.val - balanceUp;
			DurationUp = duration - balanceUp;
			DurationDown = duration + balanceDown;
			float actualIncrement = myDirection / (myDirection > 0 ? DurationUp : DurationDown);
			myValue = Mathf.Clamp01(myValue + actualIncrement * Time.deltaTime);

			TriggerUpdate();

			float power = myPower.val * myPowerScale / POWER_SCALE_MAX;
			float v = myValue;
			if (myAPCompensation.val)
				v = 1.0f-Mathf.Pow(1.0f-v, 1.6f-1.2f*Mathf.Abs(power-0.5f));
			if (mySmooth.val)
				v = Mathf.SmoothStep(0.0f, power, v);
			else
				v = power*v;

			myFloatTrigger.Trigger(v);

			if (myDirection > 0.0f && myValue < 1.0f)
				return;
			if (myDirection < 0.0f && myValue > 0.0f)
				return;


			myDirection = -myDirection;
			bool isDown = myValue <= 0.0f;

			TriggerReset();

			// gradually increase/decrease multiplier for smoother transition
			int multiplierValue = Mathf.RoundToInt(myMultiplier.val);
			if (multiplierValue < myMultiplierValue)
				--myMultiplierValue;
			else if (multiplierValue > myMultiplierValue)
				++myMultiplierValue;

			// compute required velocity for sync
			float available = myThrustEvent.Time - myBreathing.CurrentTime;
			float count = available * 2.0f * myMultiplierValue / myThrustEvent.Duration;
			if (isDown)
				count = Mathf.Round(count*0.5f+0.5f)*2.0f-1.0f; // round to odd
			else
				count = Mathf.Round(count*0.5f)*2.0f; // round to even
			float powerScale = myPowerScale;
			float velocity = (count>=0.5f) ? (count/available) : (2.0f*myMultiplierValue/myThrustEvent.Duration);
			float factor = velocity / (myIncrement*myPowerScale);

			float vc = myBreathing.VelocityClampScale;

			// partially compensate velocity difference by varying power (=range of motion)
			float vftmin = VELOCITY_FACTOR_THRESHOLD_MIN*vc;
			float vftmax = VELOCITY_FACTOR_THRESHOLD_MAX*vc;
			if (isDown && factor < vftmin)
				powerScale = Mathf.Max(factor/vftmin, POWER_SCALE_MIN);
			else if (isDown && factor > vftmax)
				powerScale = Mathf.Min(factor/vftmax, POWER_SCALE_MAX);

			// hard clamp velocity difference, compute new increment
			factor = Mathf.Clamp(factor, VELOCITY_FACTOR_MIN*vc, VELOCITY_FACTOR_MAX*vc);
			myIncrement = factor * myPowerScale/powerScale * myIncrement;
			myPowerScale = powerScale;
		}

		private const float POWER_SCALE_MIN = 0.7f;
		private const float POWER_SCALE_MAX = 1.2f;
		private const float VELOCITY_FACTOR_THRESHOLD_MIN = 0.85f;
		private const float VELOCITY_FACTOR_THRESHOLD_MAX = 1.15f;
		private const float VELOCITY_FACTOR_MIN = 0.66f;
		private const float VELOCITY_FACTOR_MAX = 1.50f;

		private void TriggerUpdate()
		{
			float signedIncrement = myIncrement*myDirection;
			myEventUp = Mathf.Max(1.0f - myValue, 0.0f) / -signedIncrement;
			myEventDown = Mathf.Max(myValue, 0.0f) / signedIncrement;
			for (int i=0; i<myTriggers.Count; ++i)
			{
				Trigger t = myTriggers[i];
				if (t.myWasTriggered)
					continue;

				t.myWasTriggered = t.myTriggerType == TRG_Down && myEventDown >= t.myOffset
				                || t.myTriggerType == TRG_Up   && myEventUp   >= t.myOffset;
				if (t.myWasTriggered)
					t.myCallback();
			}
		}

		private void TriggerReset()
		{
			int triggerTypeToReset = (myValue < 0.5f) ? TRG_Up : TRG_Down;
			for (int i=0; i<myTriggers.Count; ++i)
			{
				Trigger t = myTriggers[i];
				if (t.myTriggerType == triggerTypeToReset)
					t.myWasTriggered = false;
			}
		}

		private class ThrustEvent : MacGruber.Breathing.Event
		{
			public override float Time { get { return Marker.Time; } }
			public float Offset { get { return Marker.MarkerOffset; } set { Marker.MarkerOffset = value; } }
			public float Duration;
			private SimpleMarker Marker;
			private DriverThrust Driver;

			public ThrustEvent(DriverThrust aDriver)
			{
				Marker = SM(MRK_BreathOut);
				Driver = aDriver;
				Duration = 3.0f;
			}

			internal override void Reset()
			{
				Marker.Reset();
			}

			internal override void Enqueue(Breathing aBreathing)
			{
				Marker.Enqueue(aBreathing);
			}

			internal override void Invalidate(Breathing aBreathing)
			{
				Marker.Invalidate(aBreathing);
			}

			internal override void Update(Breathing aBreathing)
			{
				if (aBreathing.CurrentTime >= Marker.Time)
				{
					float previousTime = Marker.Time;
					Marker.Advance(aBreathing);
					Duration = Marker.Time - previousTime;
				}
				Driver.ThrustUpdate();
			}
		}

		public static readonly int TRG_Invalid = -1;
		public static readonly int TRG_Down = 0;
		public static readonly int TRG_Up = 1;

		public class Trigger
		{
			public int myTriggerType = TRG_Invalid;
			public float myOffset = 0.0f;
			public CallbackVoid myCallback;
			public bool myWasTriggered = false;
		}

		public Trigger RegisterTrigger(int triggerType, float offset, CallbackVoid callback)
		{
			Trigger t = new Trigger() {
				myTriggerType = triggerType,
				myOffset = offset,
				myCallback = callback
			};
			myTriggers.Add(t);
			return t;
		}

		public void UnregisterTrigger(Trigger trigger)
		{
			myTriggers.Remove(trigger);
		}

		private List<Trigger> myTriggers = new List<Trigger>();
	}
}