using UnityEngine;
using System;
using System.Collections.Generic;
using static MacGruber.DriverThrust;
using static MacGruber.Utils;
using SimpleJSON;

namespace MacGruber
{
	public class DriverThrustTrigger : MVRScript
	{
		private DriverThrust myDriver;

		private JSONStorableStringChooser myEventChooser;

		private JSONStorableString myEventName;
		private JSONStorableStringChooser myDirection;
		private JSONStorableFloat myOffsetRatio;
		private JSONStorableFloat myOffsetFixed;
		private JSONStorableFloat myVelocityMin;
		private JSONStorableFloat myVelocityMax;

		private List<object> myEventUI = new List<object>();
		private List<Event> myEvents = new List<Event>();
		private bool myNeedUIRefresh = false;

		private readonly List<string> DIRECTIONS = new List<string>() { "UP", "DOWN" };

		private class Event
		{
			public Event(DriverThrustTrigger script, string name)
			{
				myName = name;
				ValueTrigger = new FloatTrigger(script, "On" + name);
				ThrustTrigger = script.myDriver.RegisterTrigger(TRG_Down, 0.0f, () => script.OnTrigger(this));
			}

			public static int SortByNameAscending(Event a, Event b)
			{
				return a.Name.CompareTo(b.Name);
			}

			public string Name {
				get { return myName; }
				set {
					myName = value;
					ValueTrigger.Name = "On" + value;
				}
			}

			public bool Down {
				get { return myDown; }
				set {
					myDown = value;
					ThrustTrigger.myTriggerType = myDown ? TRG_Down : TRG_Up;
				}
			}

			public float OffsetRatio = 0.0f;
			public float OffsetFixed = 0.0f;
			public float VelocityMin = 0.1f;
			public float VelocityMax = 3.0f;
			public FloatTrigger ValueTrigger;
			public DriverThrust.Trigger ThrustTrigger;
			private string myName;
			private bool myDown = true;
		}


		public override void Init()
		{
			Utils.OnInitUI(CreateUIElement);

			if (containingAtom.type != "Person")
			{
				SuperController.LogError("MacGruber DriverThrustTrigger script needs to be on a Person atom.");
				return;
			}

			myDriver = FindWithinSamePlugin<DriverThrust>(this);
			if (myDriver == null)
			{
				SuperController.LogError("Plugin 'MacGruber.DriverThrust' not found.");
				return;
			}

			Utils.SetupInfoTextNoScroll(this,
				"<color=#606060><size=40><b>DriverThrustTrigger</b></size>\nSetup <i>Events</i> to trigger things in sync with <i>DriverThrust</i> plugin. Also provides a velocity value at the time of trigger.\n\n" +
				"<b>Name:</b> Rename your event as you like.\n\n" +
				"<b>Direction:</b> Whether the event is relative to the up or down position of thrust.\n\n" +
				"<b>Offset Ratio:</b> Time offset as ratio of the movement duration. For example a negative value of -0.2 indicates the event would be triggered earlier, at 80% of the way toward the position set as <i>Direction</i>. Similar a positive value of 0.4 indicates the event would be triggered after, at 40% of the way away from the position set as <i>Direction</i>.\n\n" +
				"<b>Offset Fixed:</b> Time offset in seconds. For example audio has often a bit of silence at the beginning, so you might want to trigger slightly earlier using a negative value here.\n\n" +
				"<b>Velocity Min/Max:</b> Thrust velocity between <i>Velocity Min</i> and <i>Velocity Max</i> are mapped to a trigger value in the 0 to 1 range. For example this can be used to adjust audio volume based on velocity.\n\n" +
				"<b>Actions:</b> Define what kinds of actions happen when the event is triggered.</color>\n\n",
				1150.0f, true
			);

			myEventChooser = Utils.SetupStringChooser(this, "Event", new List<string>(), false);
			myEventChooser.isStorable = myEventChooser.isRestorable = false;
			myEventChooser.setCallbackFunction += UISetEvent;

			Utils.SetupButton(this, "Add Event", UIAddEvent, false);
			Utils.SetupButton(this, "Remove Event", UIRemoveEvent, false);

			Utils.SetupSpacer(this, 15, false);

			myEventName = new JSONStorableString("Name", "", UIRenameEvent);
			myDirection = new JSONStorableStringChooser("Direction", DIRECTIONS, DIRECTIONS[1], "Direction");
			myDirection.setCallbackFunction += UISetDirection;
			myOffsetRatio = new JSONStorableFloat("Offset Ratio", 0.0f, -1.0f, 1.0f, true, true);
			myOffsetRatio.setCallbackFunction += UISetOffsetRatio;
			myOffsetFixed = new JSONStorableFloat("Offset Fixed", 0.0f, -0.5f, 0.5f, true, true);
			myOffsetFixed.setCallbackFunction += UISetOffsetFixed;
			myVelocityMin = new JSONStorableFloat("Velocity Min", 0.1f, 0.1f, 5.0f, true, true);
			myVelocityMin.setCallbackFunction += UISetVelocityMin;
			myVelocityMax = new JSONStorableFloat("Velocity Max", 3.0f, 0.1f, 5.0f, true, true);
			myVelocityMax.setCallbackFunction += UISetVelocityMax;

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void UISetEvent(string v)
		{
			Utils.RemoveUIElements(this, myEventUI);
			Utils.OnInitUI(CreateUIElement);

			int idx = myEventChooser.choices.IndexOf(v);
			if (idx < 0)
				return;

			Event ev = myEvents[idx];

			JSONStorableString header = Utils.SetupInfoOneLine(this, "<size=30><b>Event</b></size>", false);
			myEventUI.Add(header);

			myEventName.valNoCallback = ev.Name;
			UIDynamicLabelInput nameInput = Utils.SetupTextInput(this, "Name", myEventName, false);
			myEventUI.Add(nameInput);

			myDirection.valNoCallback = (ev.Down ? DIRECTIONS[1] : DIRECTIONS[0]);
			CreatePopup(myDirection, false);
			myEventUI.Add(myDirection);

			myOffsetRatio.valNoCallback = ev.OffsetRatio;
			CreateSlider(myOffsetRatio, false);
			myEventUI.Add(myOffsetRatio);

			myOffsetFixed.valNoCallback = ev.OffsetFixed;
			CreateSlider(myOffsetFixed, false);
			myEventUI.Add(myOffsetFixed);

			myVelocityMin.valNoCallback = ev.VelocityMin;
			CreateSlider(myVelocityMin, false);
			myEventUI.Add(myVelocityMin);

			myVelocityMax.valNoCallback = ev.VelocityMax;
			CreateSlider(myVelocityMax, false);
			myEventUI.Add(myVelocityMax);

			UIDynamicButton actionsButton = Utils.SetupButton(this, "Actions " + ev.ValueTrigger.Name, ev.ValueTrigger.OpenPanel, false);
			myEventUI.Add(actionsButton);
		}

		private void UISetDirection(string v)
		{
			int idx = myEventChooser.choices.IndexOf(myEventChooser.val);
			if (idx >= 0)
				myEvents[idx].Down = (v == DIRECTIONS[1]);
		}

		private void UISetOffsetRatio(float v)
		{
			int idx = myEventChooser.choices.IndexOf(myEventChooser.val);
			if (idx >= 0)
				myEvents[idx].OffsetRatio = v;
		}

		private void UISetOffsetFixed(float v)
		{
			int idx = myEventChooser.choices.IndexOf(myEventChooser.val);
			if (idx >= 0)
				myEvents[idx].OffsetFixed = v;
		}

		private void UISetVelocityMin(float v)
		{
			int idx = myEventChooser.choices.IndexOf(myEventChooser.val);
			if (idx >= 0)
			{
				myEvents[idx].VelocityMin = v;
				myEvents[idx].VelocityMax = Mathf.Max(v, myEvents[idx].VelocityMax);
				myVelocityMax.valNoCallback = myEvents[idx].VelocityMax;
			}
		}

		private void UISetVelocityMax(float v)
		{
			int idx = myEventChooser.choices.IndexOf(myEventChooser.val);
			if (idx >= 0)
			{
				myEvents[idx].VelocityMax = v;
				myEvents[idx].VelocityMin = Mathf.Min(v, myEvents[idx].VelocityMin);
				myVelocityMin.valNoCallback = myEvents[idx].VelocityMin;
			}
		}

		private void UIAddEvent()
		{
			for (int i=1; i<1000; ++i)
			{
				string name = "Event" + i;
				if (!myEventChooser.choices.Contains(name))
				{
					myEvents.Add(new Event(this, name));
					myEvents.Sort(Event.SortByNameAscending);
					List<string> choices = myEventChooser.choices;
					choices.Clear();
					for (int e=0; e<myEvents.Count; ++e)
						choices.Add(myEvents[e].Name);
					myEventChooser.choices = null; // force UI sync
					myEventChooser.choices = choices;
					myEventChooser.val = name;
					return;
				}
			}
			SuperController.LogError("DriverThrustTrigger: Too many events!");
		}

		private void UIRemoveEvent()
		{
			int idx = myEventChooser.choices.IndexOf(myEventChooser.val);
			if (idx < 0)
				return;

			List<string> choices = myEventChooser.choices;
			choices.RemoveAt(idx);
			myEvents[idx].ValueTrigger.Remove();
			if (myDriver != null)
				myDriver.UnregisterTrigger(myEvents[idx].ThrustTrigger);
			myEvents.RemoveAt(idx);
			myEventChooser.choices = null; // force UI sync
			myEventChooser.choices = choices;

			if (idx < choices.Count)
				myEventChooser.val = choices[idx];
			else if (choices.Count > 0)
				myEventChooser.val = choices[choices.Count-1];
			else
				myEventChooser.val = "";
		}

		private void UIRenameEvent(string name)
		{
			int idx = myEventChooser.choices.IndexOf(myEventChooser.val);
			if (idx < 0)
				return;

			myEvents[idx].Name = name;
			myEvents.Sort(Event.SortByNameAscending);

			List<string> choices = myEventChooser.choices;
			choices.Clear();
			for (int e=0; e<myEvents.Count; ++e)
				choices.Add(myEvents[e].Name);
			myEventChooser.choices = null; // force UI sync
			myEventChooser.choices = choices;
			myEventChooser.val = name;
		}

		private void OnAtomRename(string oldid, string newid)
		{
			for (int i=0; i<myEvents.Count; ++i)
				myEvents[i].ValueTrigger.SyncAtomNames();
		}

		private void OnDestroy()
		{
			if (myDriver != null)
			{
				for (int e=0; e<myEvents.Count; ++e)
					myDriver.UnregisterTrigger(myEvents[e].ThrustTrigger);
				myDriver = null;
			}

			for (int i=0; i<myEvents.Count; ++i)
				myEvents[i].ValueTrigger.Remove();
			myEvents.Clear();

			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			Utils.OnDestroyUI();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;

				JSONArray elist = new JSONArray();
				for (int i=0; i<myEvents.Count; ++i)
				{
					Event ev = myEvents[i];
					JSONClass ec = new JSONClass();
					ec["Name"] = ev.Name;
					ec["Down"].AsBool = ev.Down;
					ec["OffsetRatio"].AsFloat = ev.OffsetRatio;
					ec["OffsetFixed"].AsFloat = ev.OffsetFixed;
					ec["VelocityMin"].AsFloat = ev.VelocityMin;
					ec["VelocityMax"].AsFloat = ev.VelocityMax;
					ec[ev.ValueTrigger.Name] = ev.ValueTrigger.GetJSON(base.subScenePrefix);
					elist.Add("", ec);
				}
				jc["Events"] = elist;
			}

			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				for (int i=0; i<myEvents.Count; ++i)
				{
					myEvents[i].ValueTrigger.Remove();
					if (myDriver != null)
						myDriver.UnregisterTrigger(myEvents[i].ThrustTrigger);
				}
				myEvents.Clear();

				JSONArray elist = jc["Events"].AsArray;
				for (int i=0; i<elist.Count; ++i)
				{
					JSONClass ec = elist[i].AsObject;
					Event ev = new Event(this, ec["Name"]);
					ev.Down = ec["Down"].AsBool;
					ev.OffsetRatio = ec["OffsetRatio"].AsFloat;
					ev.OffsetFixed = ec["OffsetFixed"].AsFloat;
					ev.VelocityMin = ec["VelocityMin"].AsFloat;
					ev.VelocityMax = ec["VelocityMax"].AsFloat;
					ev.ValueTrigger.RestoreFromJSON(ec, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					myEvents.Add(ev);
				}

				myEvents.Sort(Event.SortByNameAscending);
				List<string> choices = myEventChooser.choices;
				choices.Clear();
				for (int i=0; i<myEvents.Count; ++i)
					choices.Add(myEvents[i].Name);
				myEventChooser.choices = null; // force UI sync
				myEventChooser.choices = choices;

				if (choices.Count == 0)
					myEventChooser.valNoCallback = "";
				else if (!choices.Contains(myEventChooser.val))
					myEventChooser.valNoCallback = choices[0];
				myNeedUIRefresh = true;
			}
		}

		private void Update()
		{			
			if (myDriver == null)
				return;
			
			for (int i=0; i<myEvents.Count; ++i)
				myEvents[i].ValueTrigger.Update();

			if (myNeedUIRefresh)
			{
				myNeedUIRefresh = false;
				myEventChooser.setCallbackFunction(myEventChooser.val);
			}

			for (int i=0; i<myEvents.Count; ++i)
			{
				Event ev = myEvents[i];
				float durationBefore = ev.Down ? myDriver.DurationDown : myDriver.DurationUp;
				float durationAfter = ev.Down ? myDriver.DurationUp : myDriver.DurationDown;
				float offset = ev.OffsetRatio >= 0.0f ? durationAfter : durationBefore;
				offset = offset * ev.OffsetRatio + ev.OffsetFixed;
				offset = Mathf.Clamp(offset, Mathf.Min(-durationBefore+0.05f, 0.0f), Mathf.Max(durationAfter-0.05f, 0.0f));
				ev.ThrustTrigger.myOffset = offset;
			}
		}

		private void OnTrigger(Event ev)
		{
			float duration = ev.Down ? myDriver.DurationDown : myDriver.DurationUp;
			float velocity = myDriver.Power / duration;
			velocity = Mathf.InverseLerp(ev.VelocityMin, ev.VelocityMax, velocity);
			velocity = Mathf.Clamp01(velocity);
			ev.ValueTrigger.Trigger(velocity);
		}
	}
}